searchData={"items":[{"type":"module","title":"Glass","doc":"Glass is a small library which provides a way to _transparently_ proxy function calls\nin a pluggable fashion.","ref":"Glass.html"},{"type":"module","title":"Installation - Glass","doc":"Add `:glass` to the list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:glass, \"~> 0.1.0\"}\n  ]\nend\n```","ref":"Glass.html#module-installation"},{"type":"module","title":"About - Glass","doc":"This is immediately useful if you're working with multiple applications potentially deployed\nin different releases/environments but need to call functions between them.\n\nUsually, you could do this via the built-in `:rpc` or `:erpc` modules (amongst others), but\nthis requires ceremony and boilerplate to set up and maintain.\n\nGlass aims to live up to its name, and the idea that the BEAM is network-transparent, by\ncompletely hiding the fact that you're _not just calling a local function_.","ref":"Glass.html#module-about"},{"type":"module","title":"Usage - Glass","doc":"Once installed, you can use `Glass` as simply as adding the following to your code:\n\n```elixir\n# 1) Setup and initialize `Glass`\nuse Glass\n\n# 2) Define a `GlassProxy` for the target module, which for this example, we can\n#    imagine is in another Elixir application which is bundled in the same umbrella\n#    project as the current application, but is *released separately* in a production\n#    environment so *we can't directly call functions in it*.\ndefglass ReportService, via: Glass.Methods.RPC\n\n# 3) Be blown away by `Glass` letting you call these functions directly...\ndef accounts_receivable_report do\n  current_user()\n  |> MyApp.Accounting.accounts_receivable()\n  |> ReportService.to_excel!()\nend\n```","ref":"Glass.html#module-usage"},{"type":"module","title":"Configuration - Glass","doc":"Glass is designed to be pluggable and extensible, and comes with a few built-in methods\nfor proxying function calls between modules.","ref":"Glass.html#module-configuration"},{"type":"module","title":"Built-in Methods - Glass","doc":"- `Glass.Methods.Apply` (default): Useful for prototyping locally, this method simply\n  calls the target function directly in the target module via `Kernel.apply/3`.\n- `Glass.Methods.RPC`: This method uses `:rpc` to call the target function in the target\n  module. The node the rpc call is executed on is currently limited to a random node in\n  the cluster (`:erlang.nodes()`), but this may be configurable in the future.","ref":"Glass.html#module-built-in-methods"},{"type":"module","title":"Custom Methods - Glass","doc":"You can also define your own method for proxying function calls by implementing the\n`Glass.Method` behaviour. If you want to customize the way `Glass` proxies function calls,\nyou can define your own module and pass it as the `:via` option to `defglass`.\n\n```elixir\ndefmodule MyApp.Tupleize do\n  @behaviour Glass.Method\n\n  @impl Glass.Method\n  def handle_proxy(module, function, args) do\n    {module, function, args}\n  end\nend\n\ndefmodule MyApp.Users do\n  use Glass\n\n  defglass ReportService, via: MyApp.Tupleize\n\n  ...\n\n  def users_report do\n    current_org()\n    |> list_users()\n    |> Enum.map(&build_report_rows/1)\n    |> ReportService.to_csv!()\n  end\nend\n\niex(1)> MyApp.Users.users_report()\n{ReportService, :to_csv!, [ ... ]}\n```","ref":"Glass.html#module-custom-methods"},{"type":"module","title":"Additional Configuration - Glass","doc":"- `:private` (default: `true`): Whether or not to generate the proxy module as private.\n  This is useful if you want to prevent direct access to the proxy module and force all\n  calls to go through `Glass`. Mainly affects tab-completion in IEx.\n\n- `:debug` (default: `false`): Whether or not to generate debug information when proxying\n  function calls. This is useful for debugging and tracing calls between modules.","ref":"Glass.html#module-additional-configuration"},{"type":"module","title":"Caveats - Glass","doc":"- **Magic**: This is a _very_ magical library and should be used with caution. It's designed\n  to be a _transparent_ proxy, but this means that it can be difficult to debug when things\n  go wrong. Use with caution and test thoroughly.\n\n- **Performance**: The easier it is to write network-transparent code, the easier it is to\n  write _slow_ network-transparent code.\n\n  Be mindful of the performance implications of calling functions between modules in this way\n  as depending on the `:via` method used, there may be a non-trivial serialization-deserialization\n  overhead, or limitations on how much data can be sent between nodes before issues arise.\n\n- **Security**: This library is designed to be used in a trusted environment where you have\n  control over the nodes in the cluster.\n\n  It's not designed to be used in a hostile environment where you need to worry about malicious\n  actors. Be mindful of the security implications of calling functions between modules in this way.\n\n  You can mitigate some of these concerns by using the `:via` option to define your own\n  method for proxying function calls with additional security checks, but this is left as\n  an exercise to the reader.","ref":"Glass.html#module-caveats"},{"type":"module","title":"Future Work - Glass","doc":"- **Anonymous via functions**: Allow for anonymous functions to be passed as the `:via` option\n  to `defglass` to allow for more flexible proxying of function calls.\n\n- **Customizable debug output**: Allow for custom debug output to be generated when proxying\n  function calls.\n\n- **Customizable error handling**: Allow for custom error handling to be defined when proxying\n  function calls.\n\n- **Telemetry integration**: Allow for `Telemetry` events to be emitted when proxying function\n  calls to allow for better observability of function calls between modules.\n\n- **Testing utilities**: Allow for easier testing of `Glass`-proxied functions by providing\n  utilities to mock out the `Glass` proxying mechanism.","ref":"Glass.html#module-future-work"},{"type":"module","title":"License - Glass","doc":"`Glass` is released under the MIT License. See the [LICENSE](LICENSE) file for more information.","ref":"Glass.html#module-license"},{"type":"macro","title":"Glass.defglass/2","doc":"Defines a new Glass proxy module for the given module alias.","ref":"Glass.html#defglass/2"},{"type":"macro","title":"Examples - Glass.defglass/2","doc":"Assuming you have a module `ReportingService` in another application `MyOtherApp`, which\nis deployed in another release in a different node in your cluster, you can't directly\ncall functions in that module from your current application.\n\nBy defining a Glass proxy module, you can transparently call functions in `ReportingService`,\nin this case, via `:rpc`.\n\n    defglass ReportingService, via: Glass.Methods.RPC\n\nFollowing this, all calls to `ReportingService` will be transparently proxied via `:rpc`,\nbut these appear to the reader as completely normal, local function calls.\n\nAdditionally, you can add debug logging to your Glass proxy functions to aide in debugging:\n\n    defglass ReportingService, via: Glass.Methods.Apply, debug: true\n\nAnd you can also make the proxy module public, if you need to access it directly or via\nthe REPL.\n\n    defglass ReportingService, via: Glass.Methods.Apply, private: false","ref":"Glass.html#defglass/2-examples"},{"type":"macro","title":"Options - Glass.defglass/2","doc":"* `:via` - The method to use for calling proxied functions. Defaults to `Glass.Methods.Apply`.\n* `:private` - Whether or not to generate the proxy module as private. Defaults to `true`.\n* `:debug` - Whether or not to generate proxy debug information. Defaults to `false`.","ref":"Glass.html#defglass/2-options"},{"type":"macro","title":"See Also - Glass.defglass/2","doc":"* The `Glass.Method` behaviour, which is the interface that must be matched for any `:via` method.\n* The `Glass.Methods.Apply` method, which is the simplest possible method you can implement.\n* The `Glass.Methods.RPC` method, which is a more complex method that uses `:rpc` to call functions.","ref":"Glass.html#defglass/2-see-also"},{"type":"behaviour","title":"Glass.Method","doc":"Behavior for implementing custom handling of proxying of functions in `Glass` proxies.\n\nCurrently, only the `handle_proxy/3` callback is required to be implemented.\nNo other callbacks can be implemented at this time.\n\nWhen a function on a `Glass` proxy is called, the proxied module name, function being executed,\nand arguments provided are passed to the `handle_proxy/3` callback.\n\nAny return value from the callback will be returned as the result of the function call on the proxy.\n\nFor more concrete examples, see the provided modules under the `Glass.Methods` namespace.","ref":"Glass.Method.html"},{"type":"callback","title":"Glass.Method.handle_proxy/3","doc":"","ref":"Glass.Method.html#c:handle_proxy/3"},{"type":"module","title":"Glass.Methods.Apply","doc":"This module is a `Glass.Method` implementation that simply calls the function directly via `Kernel.apply/3`.\n\n> #### Note {: .info}\n> Note that this module is just a simple example. `Kernel.apply/3` cannot be used to call > functions on\n> remote nodes, so this module is only useful for local development and testing of `Glass` itself.","ref":"Glass.Methods.Apply.html"},{"type":"module","title":"Glass.Methods.RPC","doc":"This module is a `Glass.Method` implementation that uses `:rpc` to call functions.\n\nThis is useful for calling functions on remote nodes transparently via `Glass` proxies.\n\nNote that the node to call the function on is chosen randomly from the list of nodes available\nin your Erlang cluster. If you need more control over which node to call, you should implement\nyour own `Glass.Method` module based on this one and use that instead.\n\n> #### Warning {: .warning}\n> By default, `:rpc` calls allow for remote calling of *any* function, in *any* module, > on the\n> remote node. This can be a security risk if you are not careful.\n>\n> If you need to restrict the functions that can be called remotely, you should implement your own\n> `Glass.Method` module based on this one and use that instead, making sure to implement your own\n> security checks as needed.","ref":"Glass.Methods.RPC.html"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,50,46,50]}}